#include "Includes.h"
//#include <chrono>
#include <direct.h>
#include <stdlib.h>
#include <stdio.h>

// Project includes
#include "dVKinematics.h"

GLchar* VREP_VS =
"#version 450 core\n \
layout(location=0) in vec3 in_vertex;\n \
layout(location=1) in vec2 in_texCoord;\n \
uniform float hit; \n \
uniform uint isLeft; \n \
out vec2 b_coordTexture; \n \
void main()\n \
{\n \
	if (isLeft == 1U)\n \
	{\n \
		b_coordTexture = in_texCoord;\n \
		gl_Position = vec4(in_vertex.x - hit, in_vertex.y, in_vertex.z,1);\n \
	}\n \
	else \n \
	{\n \
		b_coordTexture = vec2(1.0 - in_texCoord.x, in_texCoord.y);\n \
		gl_Position = vec4(-in_vertex.x + hit, in_vertex.y, in_vertex.z,1);\n \
	}\n \
}";

GLchar* VREP_FS =
"#version 450 core\n \
uniform sampler2D u_textureVREP; \n \
in vec2 b_coordTexture;\n \
out vec4 out_color; \n \
void main()\n \
{\n \
	out_color = vec4(texture(u_textureVREP, b_coordTexture).rgb,1); \n \
}";

// ------------------------------------------------------da multi-devices (chai3d)-------------------------------------------------------

//------------------------------------------------------------------------------
// DECLARED VARIABLES
//------------------------------------------------------------------------------

//a haptic device handler
cHapticDeviceHandler* handler;

// maximum number of devices supported by this application
const int MAX_DEVICES = 16;

// a pointer to the current haptic device
cGenericHapticDevicePtr hapticDevice[MAX_DEVICES];

// number of haptic devices detected
int numHapticDevices = 0;

// labels to display each haptic device model
cLabel* labelHapticDeviceModel[MAX_DEVICES];

// labels to display the position [m] of each haptic device
cLabel* labelHapticDevicePosition[MAX_DEVICES];

// global variable to store the position [m] of each haptic device
cVector3d hapticDevicePosition[MAX_DEVICES];

// some small spheres (cursor) representing position of each haptic device
cShapeSphere* cursor[MAX_DEVICES];

// some lines representing the velocity vector of each haptic device
cShapeLine* velocity[MAX_DEVICES];

// a flag to indicate if the haptic simulation currently running
bool simulationRunning = false;

// a flag to indicate if the haptic simulation has terminated
bool simulationFinished = true;

// a frequency counter to measure the simulation graphic rate
cFrequencyCounter freqCounterGraphics;

// a frequency counter to measure the simulation haptic rate
cFrequencyCounter freqCounterHaptics;

// haptic thread
cThread* hapticsThread;

// Shared Oculus 

// --------------------------------------------------------------
// VARAIABLES FOR THE "MOUSE EFFECT"
float haptic_position_x = 0;
float haptic_position_y = 0;
float haptic_position_z = 0;
float haptic_position_xOLD = 0;
float haptic_position_yOLD = 0;
float haptic_position_zOLD = 0;

//GET POSITION HAPTIC
float px;
float py;
float pz;
float get_haptic_position[3] = { px, py, pz };
float* get_haptic_positionPtr = get_haptic_position;

// SET POSITION HAPTIC
float set_px;
float set_py;
float set_pz;
//float set_haptic_position[3];
float* set_haptic_positionPtr;

//GET VELOCITY (LINEAR) of HAPTIC from VREP 
float vlx;
float vly;
float vlz;
float vrep_linear_velocity[3] = { vlx, vly, vlz };
float *vrep_linear_velocityPtr = vrep_linear_velocity;

float vmax = 2.5;
float wmax = 2.5;
// --------------------------------------------------------------

//------------------------------------------------------------------------------
// DECLARED FUNCTIONS
//------------------------------------------------------------------------------

// this function contains the main haptics simulation loop
void hapticCHAICallback(void);
//void hapticOpenCallback(void);
//HDCallbackCode HDCALLBACK HDHapticCallback(void *data);

void enableMovement(const bool& button, bool& button_prev, bool& raise, bool& trail, bool& trigger);
void saveToFile(std::stringstream& buff, std::string name);
void initLog();
void logData();

Eigen::Matrix3f eulAng2Rot(const float& alpha, const float& beta, const float& gamma);
Eigen::Vector3f projVec(const Eigen::Vector3f& v, const Eigen::Vector3f& p);

//==============================================================================

// --------------------------------------------------------------------------------------------------------------------------------------

// Log variables
char curDirAndFile[1024];
std::string path;
int fileSeq;

// Oculus stringstream
std::stringstream vtrSS;				//!< 6D Velocity of the rift display expressed in the tracker reference frame
std::stringstream vrrSS;				//!< 6D Velocity of the rift display expressed in the rift reference frame
std::stringstream Ttr;					//!< Homogeneous transformation matrix expressing the pose of the Oculus rift wrt the tracker reference frame

// ECM stringstream
std::stringstream qdotECMSS;			//!< 4x1 Vector of ECM joint velocities
std::stringstream qECMSS;				//!< 4x1 Vector of ECM joint position
std::stringstream TbcSS;				//!< Homogeneous transformation matrix expressing the pose of the left camera wrt the base frame of the ECM
std::stringstream vcamLSS;				//!< Velocity vector of the left camera of the ECM
std::stringstream vcamRSS;				//!< Velocity vector of the right camera of the ECM

// Geomagic stringstream
std::stringstream leftHipBaseVelSS;		//!< 6D Velocity of the left hip with respect to the base frame of the left geomagic
std::stringstream rightHipBaseVelSS;	//!< 6D Velocity of the right hip with respect to the base frame of the right geomagic
std::stringstream leftHipPSMVelSS;		//!< 6D Velocity of the left hip with respect to the base frame of the Left PSM
std::stringstream rightHipPSMVelSS;		//!< 6D Velocity of the right hip with respect to the base frame of the Left PSM
std::stringstream leftHipRotSS;			//!< Rotation matrix expressing the orientation of the left hip wrt the base frame of the left geomagic
std::stringstream rightHipRotSS;		//!< Rotation matrix expressing the orientation of the left hip wrt the base frame of the right geomagic
std::stringstream leftHipPosSS;			//!< Position of the left hip wrt the base frame of the left geomagic
std::stringstream rightHipPosSS;		//!< Position of the right hip wrt the base frame of the right geomagic

// PSM stringstream
std::stringstream qdotPSM1SS;			//!< 6x1 Vector of PSM1 joint velocity
std::stringstream qdotPSM2SS;			//!< 6x1 Vector of PSM2 joint velocity
std::stringstream qPSM1SS;				//!< 6x1 Vector of PSM1 joint position
std::stringstream qPSM2SS;				//!< 6x1 Vector of PSM2 joint position
std::stringstream Tb1gSS;				//!< Homogeneous transformation matrix expressing the pose of the left gripper (PSM1) wrt the base frame of the PSM1
std::stringstream Tb2gSS;				//!< Homogeneous transformation matrix expressing the pose of the right gripper (PSM2) wrt the base frame of the PSM2
std::stringstream vb1gSS;				//!< Velocity vector of the left gripper (PSM1) expressed wrt base frame of the PSM1
std::stringstream vb2gSS;				//!< Velocity vector of the right gripper (PSM2) expressed wrt base frame of the PSM2

// Signals
std::stringstream clutchButtonSS;
std::stringstream holdhButtonSS;
std::stringstream leftRaisingClutchEdgeSS;
std::stringstream rightRaisingClutchEdgeSS;
std::stringstream leftTrailingClutchEdgeSS;
std::stringstream rightTrailingClutchEdgeSS;
std::stringstream leftTriggerClutchButtonSS;
std::stringstream rightTriggerClutchButtonSS;


int clientID;
simxFloat Ts;

simxInt camera;
simxInt eyes[2];

#ifdef TEST
	simxInt leftBlock, rightBlock;
#endif

bool running;

vector<float> virt_dts;
vector<float> virt_l_vels;
vector<float> virt_as;


float lowPassFrequency(float v_new, float v) {
	float output;
	double RC = 1.0 / (CUTOFF * 2 * 3.14);
	double dt = 1.0 / SAMPLE_RATE;
	double alpha = dt / (RC + dt);
	output = v + (alpha*(v_new - v));
	return output;
}

float clamp(const float& in, const float& min, const float& max) {

	return ((in < min) ? min : ((in > max) ? max : in));

}

cVector3d clampVec(const cVector3d& in, const float& min, const float& max) {

	cVector3d out;

	for (int i = 0; i < 3; i++) {
		out(i) = clamp(in(i), min, max);
	}
	
	return out;

}

bool vecExceeded(const cVector3d& in) {

	return (std::fabs(in(0)) == vmax || std::fabs(in(1)) == vmax || std::fabs(in(2)) == vmax);

}


void enableMovement(const bool& button, bool& button_prev, bool& raise, bool& trail, bool& trigger) {

	if (button && !button_prev) {
		raise = true;
	}
	if (!button && button_prev) {
		trail = true;
	}
	if (raise && trail) {

		trigger = !trigger;
		raise = false;
		trail = false;
	}

	button_prev = button;
	
}

Eigen::Vector3f projVec(const Eigen::Vector3f& v, const Eigen::Vector3f& p) {

	Eigen::Vector3f ret;
	float scalar;

	scalar = v.dot(p);
	scalar /= p.norm();

	ret = scalar * p;

	return ret;

}
// --------------------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------------------------------------------

void hapticCHAICallback() {

#ifdef TEST
	simxInt blocksH[ARMS];
	simxFloat blockVel[ARMS][LIN_VEL_DIM];
	const std::string blockNames[] = { "ConcretBlock_left", "ConcretBlock_right" };
#endif

	simxInt qHandlers[ARMS][PSM_JOINTS_NUM];
	Matrix6f J_be[ARMS];
	Vector6f q6[ARMS], q6dot[ARMS], q6cmd[ARMS];
	Vector7f qcmd[ARMS], q[ARMS], qdot[ARMS]; // Vector of joint position taking into account the pair of gripper actuation (J6 = f(q6,q7))
	Vector6f hipVel[ARMS], gripperVel[ARMS], gripperVel2[ARMS], resVel[ARMS];
	cVector3d hipPos[ARMS];
	cVector3d hipLinVel[ARMS];
	cVector3d hipAngVel[ARMS];
	cMatrix3d Rbh[ARMS]; // Rotation matrix from base frame to hip frame
	cVector3d f_att[ARMS], f_rep[ARMS]; // Attractive and repulsive force
	Eigen::Matrix3f Rwb[ARMS]; // Rotation matrix from the V-REP world frame to the base frame (RCM) of the PSM
	Eigen::Matrix3f Rwb2[ARMS]; // Rotation matrix from the V-REP world frame to the new rotated base frame (RCM) of the PSM
	Eigen::Vector3f nb[ARMS]; // Unit normal vector of the table plane in the PSM frames
	Eigen::Vector3f projVel[ARMS];

	// Rotation matrix expressing the haptic base frame in the PSM base frame
	Eigen::Matrix3f Rmirror; 

	// Rotation matrix to rotate the base frame of the geomagic of pi/2 around y, to match the view seen from the Oculus Rift (approximation)
	Eigen::Matrix3f Rbp2;
	Eigen::Matrix3f Rbp2T;


	Matrix6f R6, R6g;
	LARGE_INTEGER rate, tic, toc, tac, start_t, cur_t;
	std::string qNames[ARMS][PSM_JOINTS_NUM] = { { "J1_PSM1", "J2_PSM1", "J3_PSM1", "J1_TOOL1", "J2_TOOL1", "J3_dx_TOOL1", "J3_sx_TOOL1" },
												 { "J1_PSM2", "J2_PSM2", "J3_PSM2", "J1_TOOL2", "J2_TOOL2", "J3_dx_TOOL2", "J3_sx_TOOL2" } };	
	std::string velSigNames[ARMS][VEL_DIM] = { { "psm1_vx", "psm1_vy", "psm1_vz", "psm1_wx", "psm1_wy", "psm1_wz" },
										       { "psm2_vx", "psm2_vy", "psm2_vz", "psm2_wx", "psm2_wy", "psm2_wz" } };
	std::string qdotSigNames[ARMS][PSM_JOINTS_NUM-1] = { { "qLdot_1", "qLdot_2", "qLdot_3", "qLdot_4", "qLdot_5", "qLdot_6" },
									                     { "qRdot_1", "qRdot_2", "qRdot_3", "qRdot_4", "qRdot_5", "qRdot_6" } };
	
	std::string gripHoldSigNames[ARMS] = { "LGripperHold","RGripperHold" };
	std::string canHoldSigNames[ARMS] = { "canHoldLeft","canHoldRight" };
	std::string distTableValNames[ARMS] = { "distTableLVal","distTableRVal" };
	std::string RCMNames[ARMS] = {"RCM_PSM1","RCM_PSM2"};
	simxInt RCMHandlers[ARMS];
	simxInt canHoldVal[ARMS];
	simxFloat eulAng[ARMS][ANG_VEL_DIM];
	simxFloat toolLV[ARMS][3];
	simxFloat toolAV[ARMS][3];
	float Kd = 20;
	float distTableVal[ARMS];
	float tableThresh;
	float K_s = 300.0; //stiffness gain
	float K_v = 2.5; //viscous gain
	float fx_rep = 0.0;
	double dt, et, t_;
	bool clutchButton[ARMS], clutchButtonPrev[ARMS];
	bool raisingClutchEdge[ARMS], trailingClutchEdge[ARMS];
	bool triggerClutchButton[ARMS];

	bool holdButton[ARMS], holdButtonPrev[ARMS];
	bool raisingHoldEdge[ARMS], trailingHoldEdge[ARMS];
	bool triggerHoldButton[ARMS];

	std::cout << "hapticCHAICallback ... " << std::endl;
	

	//----------------------------------
	// REMINDER: LEFT -> 0 , RIGHT -> 1
	//----------------------------------


	// Get the clock rate
	QueryPerformanceFrequency(&rate);


	//----------------------------------
	// V-REP INITIALIZATION
	//----------------------------------

	// Rotation matrix to mirror the movements with respect to the view of V-REP (we assume manipulation on the robot side)
	Rmirror << 0, 1, 0,
			  -1, 0, 0,
			   0, 0, 1;//*/

	Rbp2 << 1,  0, 0,
		    0,  0, 1,
		    0, -1, 0;//*/

	/*Rbp2 << 1,  0, 0,
			0,  0, -1,
			0, 1, 0;//*/

	Rbp2T = Rbp2.transpose();


	R6.setZero();
	R6g.setZero();

	// Use this rotation matrix if you want to keep Geomagic grippers movements expressed with respect to its original reference frame
	//R6.topLeftCorner(3, 3) = Rmirror;
	//R6.bottomRightCorner(3, 3) = Rmirror;

	// Use this rotation matrix if you want to match Geomagic grippers movements with Oculus view (approximation)
	R6.topLeftCorner(3, 3) = Rmirror;
	R6.bottomRightCorner(3, 3) = Rmirror;
	R6g.topLeftCorner(3, 3) = Rbp2;
	R6g.bottomRightCorner(3, 3) = Rbp2;

	// Init clutch button
	clutchButton[LEFT] = false;
	clutchButton[RIGHT] = false;
	clutchButtonPrev[LEFT] = false;
	clutchButtonPrev[RIGHT] = false;
	raisingClutchEdge[LEFT] = false;
	raisingClutchEdge[RIGHT] = false;
	trailingClutchEdge[LEFT] = false;
	trailingClutchEdge[RIGHT] = false;
	triggerClutchButton[LEFT] = false;
	triggerClutchButton[RIGHT] = false;

	// Init hold button
	holdButton[LEFT] = false;
	holdButton[RIGHT] = false;
	holdButtonPrev[LEFT] = false;
	holdButtonPrev[RIGHT] = false;
	raisingHoldEdge[LEFT] = false;
	raisingHoldEdge[RIGHT] = false;
	trailingHoldEdge[LEFT] = false;
	trailingHoldEdge[RIGHT] = false;
	triggerHoldButton[LEFT] = false;
	triggerHoldButton[RIGHT] = false;

	distTableVal[LEFT] = 1000.0;
	distTableVal[RIGHT] = 1000.0;
	tableThresh = 0.001;

	canHoldVal[LEFT] = 0;
	canHoldVal[RIGHT] = 0;

	// Initialize joint names
	for (int i = 0; i < ARMS; i++) {

		#ifdef TEST
			// Get the handlers of the V-REP 
			simxGetObjectHandle(clientID, (blockNames[i]).c_str(), &blocksH[i], simx_opmode_blocking);
		#endif

		// Get Handlers of the RCM objects
		simxGetObjectHandle(clientID, RCMNames[i].c_str(), &RCMHandlers[i], simx_opmode_blocking);

		// Enable streaming to get the orientation of the RCM of the PSM
		simxGetObjectOrientation(clientID, RCMHandlers[i], -1, eulAng[i], simx_opmode_streaming);

		// Enable streaming to get the distances of grippers from tables
		simxGetFloatSignal(clientID, distTableValNames[i].c_str(), &distTableVal[i], simx_opmode_streaming);

		// Enable streaming to read the signal stating if you can hold something
		simxGetIntegerSignal(clientID, canHoldSigNames[i].c_str(), &canHoldVal[i], simx_opmode_streaming);

		// Initialize chai3d data structures
		hipPos[i].zero();
		hipLinVel[i].zero();
		hipAngVel[i].zero();
		Rbh[i].identity();

		// Initialize Eigen data structures
		J_be[i].setZero();
		q[i].setZero();
		q6[i].setZero();
		qdot[i].setZero();
		q6dot[i].setZero();
		qcmd[i].setZero();
		q6cmd[i].setZero();
		hipVel[i].setZero();
		gripperVel[i].setZero();
		gripperVel2[i].setZero();
		resVel[i].setZero();
		Rwb[i].setIdentity();
		nb[i].setZero();
		projVel[i].setZero();

		for (int j = 0; j < PSM_JOINTS_NUM; j++) {

			//Retrieve PSM-i joint handlers
			simxGetObjectHandle(clientID, qNames[i][j].c_str(), &qHandlers[i][j], simx_opmode_blocking);

			// Enable streaming for data buffering of PSM1 joint positions
			simxGetJointPosition(clientID, qHandlers[i][j], &q[i](j), simx_opmode_streaming);

			// Get PSM-i joint positions in blocking-call fashion (to have guarantee that a first value is captured)
			simxGetJointPosition(clientID, qHandlers[i][j], &q[i](j), simx_opmode_blocking);
			qcmd[i](j) = q[i](j);


		}



		// Set the 6D vector with coupled gripper actuation
		q6[i].topRows(PSM_JOINTS_NUM - 2) = q[i].topRows(PSM_JOINTS_NUM - 2);
		q6[i](PSM_JOINTS_NUM - 2) = 0.5 * (q[i](PSM_JOINTS_NUM - 2) - q[i](PSM_JOINTS_NUM - 1));

	}

	//----------------------------------
	// HAPTIC DEVICES INITIALIZATION
	//----------------------------------

	// create a haptic device handler
	handler = new cHapticDeviceHandler();

	// Update the haptic device handlers to get the current number of available devices
	handler->update();

	// get number of haptic devices
	numHapticDevices = handler->getNumDevices();
	std::cout << "Num devices: " << numHapticDevices << std::endl;

	// Sanity check
	if (numHapticDevices != ARMS) {
		std::cout << "ERROR: The number of devices does not coincide with the controlled manipulators. Exiting ..." << std::endl;
		running = false;
	}

	// Setup each haptic device
	for (int i = 0; i < numHapticDevices; i++)
	{
		// get a handle to the first haptic device
		handler->getDevice(hapticDevice[i], i);

		// Check availability of left/right device
		std::cout << (i ? "Left " : "Right ") << "haptic interface " << (hapticDevice[i]->isDeviceAvailable() ? "" : "NOT ") << "available. " << std::endl;

		// open a connection to haptic device
		hapticDevice[i]->open();
		std::cout << (i ? "Left " : "Right ") << "haptic interface " << (hapticDevice[i]->isDeviceReady() ? " " : "NOT ") << "connected. " << std::endl;

		// calibrate device (if necessary)
		hapticDevice[i]->calibrate();

		// if the haptic devices carries a gripper, enable it to behave like a user switch
		hapticDevice[i]->setEnableGripperUserSwitch(true);

		// Init forces to zero
		f_rep[i].zero();
		f_att[i].zero();

	}

	// Initialize time
	QueryPerformanceCounter(&start_t);



	//----------------------------------
	// MAIN LOOP
	//----------------------------------
	while (running) {
	
		// Measure starting time
		QueryPerformanceCounter(&tic);

		// do stuff ...

		for (int i = 0; i < ARMS; i++) {



			// Get swith button state
			hapticDevice[i]->getUserSwitch(0, clutchButton[i]);
			hapticDevice[i]->getUserSwitch(1, holdButton[i]);

			// Read the V-REP buffer to get the distances of grippers from tables
			simxGetFloatSignal(clientID, distTableValNames[i].c_str(), &distTableVal[i], simx_opmode_buffer);

			// Enable streaming to read the signal stating if you can hold something
			simxGetIntegerSignal(clientID, canHoldSigNames[i].c_str(), &canHoldVal[i], simx_opmode_buffer);

			// Read the V-REP buffer to get the Euler angles expressing the orientation of the RCM wrt the world frame of V-REP
			simxGetObjectOrientation(clientID, RCMHandlers[i], -1, eulAng[i], simx_opmode_buffer);

			// Build the Rotation matrix from the Euler angles
			Rwb[i] = eulAng2Rot(eulAng[i][0], eulAng[i][1], eulAng[i][2]);

			// Extract the unit normal vector
			nb[i] = (Rwb[i]).row(2);

			// Check and enable tele-operation
			enableMovement(clutchButton[i], clutchButtonPrev[i], raisingClutchEdge[i], trailingClutchEdge[i], triggerClutchButton[i]);

			// Check and enable tele-operation
			enableMovement(holdButton[i], holdButtonPrev[i], raisingHoldEdge[i], trailingHoldEdge[i], triggerHoldButton[i]);

			// Send Gripper hold state on V-REP
			simxSetIntegerSignal(clientID, gripHoldSigNames[i].c_str(), triggerHoldButton[i], simx_opmode_oneshot);
			//simxSetIntegerSignal(clientID, gripHoldSigNames[i].c_str(), holdButton[i], simx_opmode_oneshot);

			// Get position
			hapticDevice[i]->getPosition(hipPos[i]);

			// Get orientation
			hapticDevice[i]->getRotation(Rbh[i]);

			// Get linear velocity
			hapticDevice[i]->getLinearVelocity(hipLinVel[i]);

			// Get angular velocity
			hapticDevice[i]->getAngularVelocity(hipAngVel[i]);

			// Clamp vectors
			hipLinVel[i] = clampVec(hipLinVel[i], -vmax, vmax);
			//hipAngVel[i] = clampVec(hipAngVel[i], -wmax, wmax);

			// If some grasping object is close to the gripper, send a force feedback
			//std::cout << "canHoldVal[" << i << "]: " << canHoldVal[i] << std::endl;
			if (canHoldVal[i] == 1) {
				f_att[i] = { -0.5, 0.0, 0.0 };
			}
			else {
				f_att[i].zero();
			}

			#ifdef TEST

				// Convert to simxFloat
				for (int j = 0; j < LIN_VEL_DIM; j++) {
					blockVel[i][j] = hipLinVel[i](j);
				}

				// Command the blocks
				if (!vecExceeded(hipLinVel[i])) simxSetObjectPosition(clientID, blocksH[i], blocksH[i], blockVel[i], simx_opmode_oneshot);

			#endif

			// Get the PSM-i joint position from the V-REP buffer
			for (int j = 0; j < PSM_JOINTS_NUM; j++) {

				// Get the PSM-i joint position from the V-REP buffer
				simxGetJointPosition(clientID, qHandlers[i][j], &q[i](j), simx_opmode_buffer);

			}

			//Get tools velocity from V-Rep
			simxGetObjectVelocity(clientID, qHandlers[i][PSM_JOINTS_NUM - 1], toolLV[i], toolAV[i], simx_opmode_buffer);


			// Set the 6D vector with coupled gripper actuation
			q6[i].topRows(PSM_JOINTS_NUM - 2) = q[i].topRows(PSM_JOINTS_NUM - 2);
			q6[i](PSM_JOINTS_NUM - 2) = 0.5 * (q[i](PSM_JOINTS_NUM - 2) - q[i](PSM_JOINTS_NUM - 1));

			// Compute the current Jacobian matrix
			J_be[i] = PSM_Jacobian(q6[i]);

			// Build hip velocity vector
			hipVel[i].topRows(LIN_VEL_DIM) = hipLinVel[i].eigen().cast<float>();
			hipVel[i].bottomRows(ANG_VEL_DIM) = hipAngVel[i].eigen().cast<float>();

			// Convert into gripper velocity vector by multiplying for the rotation matrix Rmirror
			gripperVel[i] = R6 * hipVel[i];

			// Revert the orientation to match the view seen from the Oculus (approximation)
			gripperVel2[i] = R6g * gripperVel[i];

			if (triggerClutchButton[i]) {
			
				// Compute the projection of the linear velocity on the table plane
				projVel[i] = projVec(gripperVel2[i].topRows(LIN_VEL_DIM), nb[i]);
				fx_rep = -K_s * distTableVal[i] - K_v * projVel[i].norm();
				/*if (distTableVal[i] < tableThresh && projVel[i].dot(nb[i]) < 0) {
					std::cout << "table distance exceeded. " << std::endl;
					resVel[i].topRows(LIN_VEL_DIM) = gripperVel2[i].topRows(LIN_VEL_DIM) - projVel[i];
					Table_feedbackForce[i] = { 3.3 , 0.0, 0.0};
				}
				else {
					std::cout << "table distance NOT exceeded. " << std::endl;
					resVel[i].topRows(LIN_VEL_DIM) = gripperVel2[i].topRows(LIN_VEL_DIM);
					Table_feedbackForce[i] = { 0.0, 0.0, 0.0 };
				}
				resVel[i].bottomRows(ANG_VEL_DIM) = gripperVel2[i].bottomRows(ANG_VEL_DIM);//*/
				if (distTableVal[i] < 0) {
					f_rep[i] = { fx_rep , 0.0, 0.0 };
				}
				else {
					f_rep[i].zero();
				}
				resVel[i] = gripperVel2[i];

				// Compute the PSM-i joint velocities
				q6dot[i] = (J_be[i]).inverse() * resVel[i]; // J^-1 * v
			}
			else {
				q6dot[i].setZero();
			}

			// Compute the new PSM-i joint position vector
			//q6cmd[i] +=  q6dot[i] * Ts;

			// Apply joint velocities to PSM-i manipulator
			for (int j = 0; j < PSM_JOINTS_NUM-1; j++) {
			
				// Apply joint velocities -> Done in V-REP
				//simxSetJointTargetVelocity(clientID, qHandlers[i][j], qdot[i](j), simx_opmode_oneshot);
				//simxSetJointPosition(clientID, qHandlers[i][j], qcmd[i](j), simx_opmode_oneshot);

				// Send joint velocity data to V-REP
				simxSetFloatSignal(clientID, qdotSigNames[i][j].c_str(), q6dot[i](j), simx_opmode_oneshot);
			}

			// Send Cartesian velocity data to V-REP
			for (int j = 0; j < VEL_DIM; j++) {
				simxSetFloatSignal(clientID, velSigNames[i][j].c_str(), gripperVel2[i](j), simx_opmode_oneshot);
			}

			hapticDevice[i]->setForce(f_rep[i] + f_att[i]);
		}


		// Measure final time
		QueryPerformanceCounter(&toc);
		dt = (toc.QuadPart - tic.QuadPart) / float(rate.QuadPart);// *1000.0 * 1000.0;
		t_ = (tic.QuadPart - start_t.QuadPart) / float(rate.QuadPart);

		// Log data
		leftHipBaseVelSS << t_ << ", " << hipVel[LEFT](0) << ", " << hipVel[LEFT](1) << ", " << hipVel[LEFT](2) << ", " << hipVel[LEFT](3) << ", " << hipVel[LEFT](4) << ", " << hipVel[LEFT](5) << "; " << std::endl;
		rightHipBaseVelSS << t_ << ", " << hipVel[RIGHT](0) << ", " << hipVel[RIGHT](1) << ", " << hipVel[RIGHT](2) << ", " << hipVel[RIGHT](3) << ", " << hipVel[RIGHT](4) << ", " << hipVel[RIGHT](5) << "; " << std::endl;
		leftHipPSMVelSS << t_ << ", " << gripperVel[LEFT](0) << ", " << gripperVel[LEFT](1) << ", " << gripperVel[LEFT](2) << ", " << gripperVel[LEFT](3) << ", " << gripperVel[LEFT](4) << ", " << gripperVel[LEFT](5) << "; " << std::endl;
		rightHipPSMVelSS << t_ << ", " << gripperVel[RIGHT](0) << ", " << gripperVel[RIGHT](1) << ", " << gripperVel[RIGHT](2) << ", " << gripperVel[RIGHT](3) << ", " << gripperVel[RIGHT](4) << ", " << gripperVel[RIGHT](5) << "; " << std::endl;
		leftHipPosSS << t_ << ", " << hipPos[LEFT](0) << ", " << hipPos[LEFT](1) << ", " << hipPos[LEFT](2) << "; " << std::endl;
		rightHipPosSS << t_ << ", " << hipPos[RIGHT](0) << ", " << hipPos[RIGHT](1) << ", " << hipPos[RIGHT](2) << "; " << std::endl;

		qPSM1SS << t_ << ", " << q[LEFT](0) << ", " << q[LEFT](1) << ", " << q[LEFT](2) << ", " << q[LEFT](3) << ", " << q[LEFT](4) << ", " << q[LEFT](5) << "; " << std::endl;
		qPSM2SS << t_ << ", " << q[RIGHT](0) << ", " << q[RIGHT](1) << ", " << q[RIGHT](2) << ", " << q[RIGHT](3) << ", " << q[RIGHT](4) << ", " << q[RIGHT](5) << "; " << std::endl;
		qdotPSM1SS << t_ << ", " << hipVel[LEFT](0) << ", " << hipVel[LEFT](1) << ", " << hipVel[LEFT](2) << ", " << hipVel[LEFT](3) << ", " << hipVel[LEFT](4) << ", " << hipVel[LEFT](5) << "; " << std::endl;
		qdotPSM2SS << t_ << ", " << hipVel[RIGHT](0) << ", " << hipVel[RIGHT](1) << ", " << hipVel[RIGHT](2) << ", " << hipVel[RIGHT](3) << ", " << hipVel[RIGHT](4) << ", " << hipVel[RIGHT](5) << "; " << std::endl;
		vb1gSS << t_ << ", " << toolLV[LEFT][0] << ", " << toolLV[LEFT][1] << ", " << toolLV[LEFT][2] << ", " << toolAV[LEFT][0] << ", " << toolAV[LEFT][1] << ", " << toolAV[LEFT][2] << "; " << std::endl;
		vb2gSS << t_ << ", " << toolLV[RIGHT][0] << ", " << toolLV[RIGHT][1] << ", " << toolLV[RIGHT][2] << ", " << toolAV[RIGHT][0] << ", " << toolAV[RIGHT][1] << ", " << toolAV[RIGHT][2] << "; " << std::endl;

		// Wait if you run in less than Ts
		if (dt < Ts) { usleep((Ts - dt)*1e+6); }

		// Measure total elapsed time after waiting (sanity check)
		QueryPerformanceCounter(&tac);
		et = (tac.QuadPart - tic.QuadPart) / float(rate.QuadPart);// *1000.0 * 1000.0;
		//std::cout << "[haptic] Rate: " << 1.0 / et << std::endl << std::endl;
	}

	for (int i = 0; i < ARMS; i++) {
		for (int j = 0; j < PSM_JOINTS_NUM - 1; j++) {

			// Send joint velocity data to V-REP
			simxSetFloatSignal(clientID, qdotSigNames[i][j].c_str(), 0.0, simx_opmode_oneshot);
		}
	}

}


void escapeCallback() {
	while (_getch() != 27);
	running = false;
}

//flippare l'immagine
void riftCallback() {

	LARGE_INTEGER rate, tic, start_t;
	double t_;

	// Get the clock rate
	QueryPerformanceFrequency(&rate);

	std::cout << "Starting Rift thread\n";

	int resolution[2] = { 320, 240 };
	int status = 1;

	Mat arr[2], arr_flip[2], img, left, right;
	simxUChar * eyeBuffers[2];
	eyeBuffers[0] = new simxUChar[resolution[0] * resolution[1] * sizeof(simxUChar)];
	eyeBuffers[1] = new simxUChar[resolution[0] * resolution[1] * sizeof(simxUChar)];

	simxGetVisionSensorImage(clientID, eyes[0], resolution, &eyeBuffers[0], 0, simx_opmode_streaming);
	simxGetVisionSensorImage(clientID, eyes[1], resolution, &eyeBuffers[1], 0, simx_opmode_streaming);

	SDL_Init(SDL_INIT_VIDEO);

	ovrResult result = ovr_Initialize(nullptr);
	if (OVR_FAILURE(result)) {
		std::cout << "ERROR: Failed to initialize libOVR" << std::endl;
	}

	ovrSession session;
	ovrGraphicsLuid luid;
	result = ovr_Create(&session, &luid);
	if (OVR_FAILURE(result)) {
		std::cout << "ERROR: Oculus Rift not detected" << std::endl;
		ovr_Shutdown();
		SDL_Quit();
	}

	int x = SDL_WINDOWPOS_CENTERED, y = SDL_WINDOWPOS_CENTERED;
	int winWidth = 1280;
	int winHeight = 720;

	Uint32 flags = SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN;
	SDL_Window* window;
	window = SDL_CreateWindow("Stereo Camera", x, y, winWidth, winHeight, flags);

	SDL_GLContext glContext = SDL_GL_CreateContext(window);
	glewInit();
	SDL_GL_SetSwapInterval(0);

	GLuint naoCameraTexture;
	glGenTextures(1, &naoCameraTexture);
	glBindTexture(GL_TEXTURE_2D, naoCameraTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, resolution[0], resolution[1], 0, GL_BGRA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	ovrHmdDesc hmdDesc = ovr_GetHmdDesc(session);
	ovrSizei textureSize0 = ovr_GetFovTextureSize(session, ovrEye_Left, hmdDesc.DefaultEyeFov[0], 1.0f);
	ovrSizei textureSize1 = ovr_GetFovTextureSize(session, ovrEye_Right, hmdDesc.DefaultEyeFov[1], 1.0f);
	ovrSizei bufferSize;
	bufferSize.w = textureSize0.w + textureSize1.w;
	bufferSize.h = std::max(textureSize0.h, textureSize1.h);

	ovrTextureSwapChain textureChain = nullptr;
	ovrTextureSwapChainDesc descTextureSwap = ovrTextureSwapChainDesc();
	descTextureSwap.Type = ovrTexture_2D;
	descTextureSwap.ArraySize = 1;
	descTextureSwap.Width = bufferSize.w;
	descTextureSwap.Height = bufferSize.h;
	descTextureSwap.MipLevels = 1;
	descTextureSwap.Format = OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
	descTextureSwap.SampleCount = 1;
	descTextureSwap.StaticImage = ovrFalse;

	result = ovr_CreateTextureSwapChainGL(session, &descTextureSwap, &textureChain);

	int length = 0;
	ovr_GetTextureSwapChainLength(session, textureChain, &length);

	if (OVR_SUCCESS(result)) {
		for (int i = 0; i < length; ++i) {
			GLuint chainTexId;
			ovr_GetTextureSwapChainBufferGL(session, textureChain, i, &chainTexId);
			glBindTexture(GL_TEXTURE_2D, chainTexId);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		}
	}
	else {
		std::cout << "ERROR: failed creating swap texture" << std::endl;
		ovr_Destroy(session);
		ovr_Shutdown();
		SDL_GL_DeleteContext(glContext);
		SDL_DestroyWindow(window);
		SDL_Quit();
	}

	GLuint fboID;
	glGenFramebuffers(1, &fboID);

	GLuint depthBuffID;
	glGenTextures(1, &depthBuffID);
	glBindTexture(GL_TEXTURE_2D, depthBuffID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	GLenum internalFormat = GL_DEPTH_COMPONENT24;
	GLenum type = GL_UNSIGNED_INT;
	glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, bufferSize.w, bufferSize.h, 0, GL_DEPTH_COMPONENT, type, NULL);

	ovrMirrorTextureDesc descMirrorTexture;
	memset(&descMirrorTexture, 0, sizeof(descMirrorTexture));
	descMirrorTexture.Width = winWidth;
	descMirrorTexture.Height = winHeight;
	descMirrorTexture.Format = OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
	ovrMirrorTexture mirrorTexture = nullptr;
	result = ovr_CreateMirrorTextureGL(session, &descMirrorTexture, &mirrorTexture);
	if (!OVR_SUCCESS(result)) {
		std::cout << "ERROR: Failed to create mirror texture" << std::endl;
	}
	GLuint mirrorTextureId;
	ovr_GetMirrorTextureBufferGL(session, mirrorTexture, &mirrorTextureId);

	GLuint mirrorFBOID;
	glGenFramebuffers(1, &mirrorFBOID);
	glBindFramebuffer(GL_READ_FRAMEBUFFER, mirrorFBOID);
	glFramebufferTexture2D(GL_READ_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, mirrorTextureId, 0);
	glFramebufferRenderbuffer(GL_READ_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, 0);
	glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);

	long long frameIndex = 0;
	ovr_SetTrackingOriginType(session, ovrTrackingOrigin_FloorLevel);
	ovrPosef eyeRenderPose[2];
	ovrEyeRenderDesc eyeRenderDesc[2];
	eyeRenderDesc[0] = ovr_GetRenderDesc(session, ovrEye_Left, hmdDesc.DefaultEyeFov[0]);
	eyeRenderDesc[1] = ovr_GetRenderDesc(session, ovrEye_Right, hmdDesc.DefaultEyeFov[1]);
	ovrVector3f hmdToEyeOffset[2];
	double sensorSampleTime;

	Shader shader(VREP_VS, VREP_FS);
	float vrepFovH = (float)(60.0 * M_PI) / 180;
	float ovrFovH = (atanf(hmdDesc.DefaultEyeFov[0].LeftTan) + atanf(hmdDesc.DefaultEyeFov[0].RightTan));
	unsigned int usefulWidth = resolution[0] * ovrFovH / vrepFovH;
	unsigned int widthFinal = bufferSize.w / 2;
	float heightGL = 1.f;
	float widthGL = 1.f;

#ifndef FULLSCREEN
	if (usefulWidth > 0.f) {
		unsigned int heightFinal = resolution[1] * widthFinal / usefulWidth;
		// Convert this size to OpenGL viewport's frame's coordinates
		heightGL = (heightFinal) / (float)(bufferSize.h);
		widthGL = ((resolution[1] * (heightFinal / (float)resolution[1])) / (float)widthFinal);
	}
	else {
		std::cout << "wrong v-rep parameters" << std::endl;
	}
#endif

	float ovrFovV = (atanf(hmdDesc.DefaultEyeFov[0].UpTan) + atanf(hmdDesc.DefaultEyeFov[0].DownTan));
	float offsetLensCenterX = ((atanf(hmdDesc.DefaultEyeFov[0].LeftTan)) / ovrFovH) * 2.f - 1.f;
	float offsetLensCenterY = ((atanf(hmdDesc.DefaultEyeFov[0].UpTan)) / ovrFovV) * 2.f - 1.f;

	struct GLScreenCoordinates {
		float left, up, right, down;
	} screenCoord;
	screenCoord.up = heightGL + offsetLensCenterY;
	screenCoord.down = heightGL - offsetLensCenterY;
	screenCoord.right = widthGL + offsetLensCenterX;
	screenCoord.left = widthGL - offsetLensCenterX;

	float rectVertices[12] = { -screenCoord.left, -screenCoord.up, 0,
		screenCoord.right, -screenCoord.up, 0,
		screenCoord.right, screenCoord.down, 0,
		-screenCoord.left, screenCoord.down, 0 };
	GLuint rectVBO[3];
	glGenBuffers(1, &rectVBO[0]);
	glBindBuffer(GL_ARRAY_BUFFER, rectVBO[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(rectVertices), rectVertices, GL_STATIC_DRAW);

	float rectTexCoord[8] = { 0, 1, 1, 1, 1, 0, 0, 0 };
	glGenBuffers(1, &rectVBO[1]);
	glBindBuffer(GL_ARRAY_BUFFER, rectVBO[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(rectTexCoord), rectTexCoord, GL_STATIC_DRAW);

	unsigned int rectIndices[6] = { 0, 1, 2, 0, 2, 3 };
	glGenBuffers(1, &rectVBO[2]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, rectVBO[2]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(rectIndices), rectIndices, GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	float hit = 0.02f;
	SDL_Event events;

	unsigned int riftc = 0, zedc = 1;
	unsigned int rifttime = 0;
	int time1 = 0, timePerFrame = 0;
	int frameRate = (int)(1000 / MAX_FPS);

	glUseProgram(shader.getProgramId());
	glEnableVertexAttribArray(Shader::ATTRIB_VERTICES_POS);
	glBindBuffer(GL_ARRAY_BUFFER, rectVBO[0]);
	glVertexAttribPointer(Shader::ATTRIB_VERTICES_POS, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, rectVBO[2]);
	glEnableVertexAttribArray(Shader::ATTRIB_TEXTURE2D_POS);
	glBindBuffer(GL_ARRAY_BUFFER, rectVBO[1]);
	glVertexAttribPointer(Shader::ATTRIB_TEXTURE2D_POS, 2, GL_FLOAT, GL_FALSE, 0, 0);

	std::cout << "Starting Rift main loop\n";

	float rift[3] = { 0, 0, 0 };
	float cam_zero[3] = { 0, 0, 0 };

	int n = 0;
	float* mat = new float[16]();

	ovr_RecenterTrackingOrigin(session);

	while (true) {
		while (simxGetObjectOrientation(clientID, camera, -1, cam_zero, simx_opmode_blocking));
		int ret = 1;
		for (int i = 0; i < 3; i++) if (isnan(cam_zero[i]) || fabsf(cam_zero[i]) > M_PI) ret = 0;
		if (ret) break;
	}

	//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	simxInt J1_ECM, J2_ECM, J3_ECM, J4_ECM;

	simxGetObjectHandle(clientID, "J1_ECM", &J1_ECM, simx_opmode_blocking);
	simxGetObjectHandle(clientID, "J2_ECM", &J2_ECM, simx_opmode_blocking);
	simxGetObjectHandle(clientID, "J3_ECM", &J3_ECM, simx_opmode_blocking);
	simxGetObjectHandle(clientID, "J4_ECM", &J4_ECM, simx_opmode_blocking);

	std::cout << "J1_ECM handler: " << J1_ECM << std::endl;
	std::cout << "J2_ECM handler: " << J2_ECM << std::endl;
	std::cout << "J3_ECM handler: " << J3_ECM << std::endl;
	std::cout << "J4_ECM handler: " << J4_ECM << std::endl;

	Eigen::Matrix<float, 3, 4> jacobian;
	Eigen::Matrix<float, 3, 3> rRt;
	Eigen::Matrix<float, 3, 3> bRe;
	Eigen::Matrix3f eRr, Rflip;
	Eigen::Matrix<float, 4, 4> DirKin;
	Eigen::Matrix<float, 3, 1> p;
	Eigen::Matrix<float, 3, 1> p_vel, wrr, wee;
	Eigen::Matrix<float, 3, 1> p_vel_prec;
	Quat<float> quat;
	Eigen::Vector4f qdot;
	Eigen::Quaternionf quaternion;
	//std::vector<simxFloat> q(4, 1);
	Eigen::Vector4f q, qcmd;
	float scale = 5.0;

	double p_prec[3] = { 0, 0, 0 };
	double p_after[3];
	double p_dot[3];
	double z = 0;
	p_vel_prec.setZero();
	//auto start = std::chrono::system_clock::now();

	simxFloat data;

	// Enable streaming
	simxGetJointPosition(clientID, J1_ECM, &q(0), simx_opmode_streaming);
	simxGetJointPosition(clientID, J2_ECM, &q(1), simx_opmode_streaming);
	simxGetJointPosition(clientID, J3_ECM, &q(2), simx_opmode_streaming);
	simxGetJointPosition(clientID, J4_ECM, &q(3), simx_opmode_streaming);//*/
																		 
	// Init some data: here, we use a blocking call to be sure to read the proper values
	simxGetJointPosition(clientID, J1_ECM, &q(0), simx_opmode_blocking);
	simxGetJointPosition(clientID, J2_ECM, &q(1), simx_opmode_blocking);
	simxGetJointPosition(clientID, J3_ECM, &q(2), simx_opmode_blocking);
	simxGetJointPosition(clientID, J4_ECM, &q(3), simx_opmode_blocking);//*/
	qcmd = q;

	// Initialize time
	QueryPerformanceCounter(&start_t);

	while (running) {

		QueryPerformanceCounter(&tic);

		/*simxGetJointPosition(clientID, J1_ECM, &q[0], simx_opmode_blocking);
		simxGetJointPosition(clientID, J2_ECM, &q[1], simx_opmode_blocking);
		simxGetJointPosition(clientID, J3_ECM, &q[2], simx_opmode_blocking);
		simxGetJointPosition(clientID, J4_ECM, &q[3], simx_opmode_blocking);//*/

		// Get data from the buffer
		simxGetJointPosition(clientID, J1_ECM, &q(0), simx_opmode_buffer);
		simxGetJointPosition(clientID, J2_ECM, &q(1), simx_opmode_buffer);
		simxGetJointPosition(clientID, J3_ECM, &q(2), simx_opmode_buffer);
		simxGetJointPosition(clientID, J4_ECM, &q(3), simx_opmode_buffer);//*/

		Posef pose = ovr_GetTrackingState(session, 0, true).HeadPose.ThePose;
		pose.Rotation.GetEulerAngles<Axis_X, Axis_Z, Axis_Y>(&rift[0], &rift[1], &rift[2]);
		ovrVector3f angVel = ovr_GetTrackingState(session, 0, true).HeadPose.AngularVelocity;
		ovrVector3f linVel = ovr_GetTrackingState(session, 0, true).HeadPose.LinearVelocity;
	

		if (GetKeyState(VK_SPACE)) {

			quat = pose.Rotation;
			quaternion.x() = quat.x;
			quaternion.y() = quat.y;
			quaternion.z() = quat.z;
			quaternion.w() = quat.w;


			rRt = quaternion.toRotationMatrix();


			/*
			auto end = std::chrono::system_clock::now();
			std::chrono::duration<double> diff = end - start;
			start = end;
			auto sampling_time = diff.count();
			if (sampling_time == 0)		continue;
			*/

			if (!GetKeyState(VK_DOWN)) { //Orientation

				std::cout << "ORIENTATION CONTROL ENABLED." << std::endl;


				p(0, 0) = cam_zero[0] - rift[0];
				p(1, 0) = cam_zero[1] + rift[1];
				p(2, 0) = cam_zero[2] + rift[2];

				p_after[0] = p(0, 0);
				p_after[1] = p(1, 0);
				p_after[2] = p(2, 0);

				p_dot[0] = (p_after[0] - p_prec[0]) / Ts; // --> sampling_time
				p_dot[1] = (p_after[1] - p_prec[1]) / Ts;
				p_dot[2] = (p_after[2] - p_prec[2]) / Ts;

				data = p_dot[0];
				//simxSetFloatSignal(clientID, "before", data, simx_opmode_oneshot);

				p_vel[0] = lowPassFrequency(p_dot[0], p_vel_prec[0]);
				p_vel[1] = lowPassFrequency(p_dot[1], p_vel_prec[1]);
				p_vel[2] = lowPassFrequency(p_dot[2], p_vel_prec[2]);

				data = p_vel[0];
				//simxSetFloatSignal(clientID, "after", data, simx_opmode_oneshot);

				p_vel_prec = p_vel;

				p_prec[0] = p_after[0];
				p_prec[1] = p_after[1];
				p_prec[2] = p_after[2];//*/

				p_vel(0) = angVel.x;
				p_vel(1) = angVel.y;
				p_vel(2) = angVel.z;//*/

				wrr = rRt.transpose()*p_vel;

				// Match the rift orientation with camera frame
				eRr << 1, 0, 0,
					   0, -1, 0,
					   0, 0, -1;

				Rflip << -1,  0,  0,
						  0, -1,  0,
						  0,  0,  1;

				wee = scale * (Rflip * eRr * wrr);
				//wee = scale *  wrr;
				//wee(0) *= -1;
				//wee(1) *= -1;

				// Send data to V-REP
				simxSetFloatSignal(clientID, "ecm_wx", wee(0), simx_opmode_oneshot);
				simxSetFloatSignal(clientID, "ecm_wy", wee(1), simx_opmode_oneshot);
				simxSetFloatSignal(clientID, "ecm_wz", wee(2), simx_opmode_oneshot);

				std::cout << "angVel: " << wee.transpose() << std::endl;

				DirKin = forwardKinematicsECM(q);
				bRe = DirKin.topLeftCorner(3, 3);

				/*for (int i = 0; i < 3; i++) {
					for (int j = 0; j < 3; j++) {
						bRe(i, j) = DirKin(i, j);
					}
				}//*/

				jacobian = bRe.transpose()*ECM_Jacobian(q);
				qdot = jacobian.fullPivLu().solve(wee); // J^-1 * p
				qdot(2) = 0;

				// Send data to V-REP
				simxSetFloatSignal(clientID, "qdotECM_1", qdot(0), simx_opmode_oneshot);
				simxSetFloatSignal(clientID, "qdotECM_2", qdot(1), simx_opmode_oneshot);
				simxSetFloatSignal(clientID, "qdotECM_3", qdot(2), simx_opmode_oneshot);
				simxSetFloatSignal(clientID, "qdotECM_4", qdot(3), simx_opmode_oneshot);

				/*qcmd[0] += (qdot(0, 0) * Ts); // --> sampling_time
				qcmd[1] += (qdot(1, 0) * Ts);
				qcmd[3] += (qdot(3, 0) * Ts);//*/

			}
			else { //Zoom

				std::cout << "POSITION CONTROL ENABLED." << std::endl;
				//z = -pose.Translation.z;
				Eigen::Vector3f eigen_lin_vel;

				eigen_lin_vel[0] = linVel.x;
				eigen_lin_vel[1] = linVel.y;
				eigen_lin_vel[2] = linVel.z;

				eigen_lin_vel = rRt * eigen_lin_vel;
				
				qdot(0) = 0;
				qdot(1) = 0;
				qdot(2) = - eigen_lin_vel[2];
				qdot(3) = 0;


				//qcmd[2] += - (linVel.z * Ts); // --> sampling_time
				//simxSetFloatSignal(clientID, "z", qcmd[2], simx_opmode_oneshot);
				
			}


			// AVOID BLOCKING-CALLS!
			/*simxSetJointPosition(clientID, J1_ECM, qcmd[0], simx_opmode_blocking);
			simxSetJointPosition(clientID, J2_ECM, qcmd[1], simx_opmode_blocking);
			simxSetJointPosition(clientID, J3_ECM, z, simx_opmode_blocking);
			simxSetJointPosition(clientID, J4_ECM, qcmd[3], simx_opmode_blocking);//*/

			/*simxSetJointPosition(clientID, J1_ECM, qcmd[0], simx_opmode_oneshot);
			simxSetJointPosition(clientID, J2_ECM, qcmd[1], simx_opmode_oneshot);
			simxSetJointPosition(clientID, J3_ECM, qcmd[2], simx_opmode_oneshot);
			simxSetJointPosition(clientID, J4_ECM, qcmd[3], simx_opmode_oneshot);//*/

			//std::cout << "q: " << q[0] << std::endl;
			/*
			q[0] = temp(0, 0);
			q[1] = temp(1, 0);
			q[2] = temp(2, 0);
			q[3] = temp(3, 0);*/
		}
		else {

			qdot.setZero();
			
		}

		simxSetFloatSignal(clientID, "qdotECM_1", qdot(0), simx_opmode_oneshot);
		simxSetFloatSignal(clientID, "qdotECM_2", qdot(1), simx_opmode_oneshot);
		simxSetFloatSignal(clientID, "qdotECM_3", qdot(2), simx_opmode_oneshot);
		simxSetFloatSignal(clientID, "qdotECM_4", qdot(3), simx_opmode_oneshot);
		

		t_ = (tic.QuadPart - start_t.QuadPart) / float(rate.QuadPart);

		// Log data
		qdotECMSS << t_ << ", " << qdot(0) << ", " << qdot(1) << ", " << qdot(2) << ", " << qdot(3) << "; " << std::endl;
		qECMSS << t_ << ", " << q(0) << ", " << q(1) << ", " << q(2) << ", " << q(3) << "; " << std::endl;
		vtrSS << t_ << ", " << linVel.x << ", " << linVel.y << ", " << linVel.z << ", " << angVel.x << ", " << angVel.y << ", " << angVel.z << "; " << std::endl;
		//vrrSS << t_ << ", " << linVel.x << ", " << linVel.y << ", " << linVel.z << ", " << wrr(0) << ", " << wrr(1) << ", " << wrr(2) << "; " << std::endl;

		//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


		timePerFrame = SDL_GetTicks() - time1;

		if (timePerFrame < frameRate) {
			SDL_Delay(frameRate - timePerFrame);
			timePerFrame = frameRate;
		}

		rifttime += timePerFrame;
		riftc++;
		if (rifttime > 200) {
			rifttime = 0;
			riftc = 0;
		}
		time1 = SDL_GetTicks();

		while (SDL_PollEvent(&events)) {
			if (events.type == SDL_KEYUP) {
				if (events.key.keysym.scancode == SDL_SCANCODE_R)
					hit = 0.0f;
			}
			if (events.type == SDL_MOUSEWHEEL) {
				float s;
				events.wheel.y > 0 ? s = 1.0f : s = -1.0f;
				hit += 0.005f * s;
			}
		}

		GLuint curTexId;
		int curIndex;
		ovr_GetTextureSwapChainCurrentIndex(session, textureChain, &curIndex);
		ovr_GetTextureSwapChainBufferGL(session, textureChain, curIndex, &curTexId);

		eyeRenderDesc[0] = ovr_GetRenderDesc(session, ovrEye_Left, hmdDesc.DefaultEyeFov[0]);
		eyeRenderDesc[1] = ovr_GetRenderDesc(session, ovrEye_Right, hmdDesc.DefaultEyeFov[1]);
		hmdToEyeOffset[0] = eyeRenderDesc[0].HmdToEyeOffset;
		hmdToEyeOffset[1] = eyeRenderDesc[1].HmdToEyeOffset;
		ovr_GetEyePoses(session, frameIndex, ovrTrue, hmdToEyeOffset, eyeRenderPose, &sensorSampleTime);

		glBindFramebuffer(GL_FRAMEBUFFER, fboID);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, curTexId, 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depthBuffID, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glClearColor(0, 0, 0, 1);

		for (int i = 0; i < 2; i++) {
			simxGetVisionSensorImage(clientID, eyes[i], resolution, &eyeBuffers[i], 0, simx_opmode_buffer);
			try {
				arr[i] = Mat(resolution[1], resolution[0], CV_8UC3, eyeBuffers[i]);
				flip(arr[i], arr_flip[i], 1);
			}
			catch (cv::Exception) {
				continue;
			}

			glViewport(i == ovrEye_Left ? 0 : bufferSize.w / 2, 0, bufferSize.w / 2, bufferSize.h);
			glBindTexture(GL_TEXTURE_2D, naoCameraTexture);

			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, resolution[0], resolution[1], 0, GL_RGB, GL_UNSIGNED_BYTE, arr_flip[i].ptr());

			glUniform1f(glGetUniformLocation(shader.getProgramId(), "hit"), i == ovrEye_Left ? hit : -hit);
			glUniform1ui(glGetUniformLocation(shader.getProgramId(), "isLeft"), i == ovrEye_Left ? 1U : 0U);
			glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
		}

		glBindFramebuffer(GL_FRAMEBUFFER, fboID);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, 0, 0);
		ovr_CommitTextureSwapChain(session, textureChain);

		//frameIndex++;

		ovrLayerEyeFov ld;
		ld.Header.Type = ovrLayerType_EyeFov;
		ld.Header.Flags = ovrLayerFlag_TextureOriginAtBottomLeft;
		for (int eyes = 0; eyes < 2; ++eyes) {
			ld.ColorTexture[eyes] = textureChain;
			ld.Viewport[eyes] = OVR::Recti(eyes == ovrEye_Left ? 0 : bufferSize.w / 2, 0, bufferSize.w / 2, bufferSize.h);
			ld.Fov[eyes] = hmdDesc.DefaultEyeFov[eyes];
			ld.RenderPose[eyes] = eyeRenderPose[eyes];
		}

		ovrLayerHeader* layers = &ld.Header;
		result = ovr_SubmitFrame(session, frameIndex, nullptr, &layers, 1);

		if (!OVR_SUCCESS(result)) {
			std::cout << "ERROR: failed to submit frame" << std::endl;
			ovr_DestroyTextureSwapChain(session, textureChain);
			ovr_DestroyMirrorTexture(session, mirrorTexture);
			ovr_Destroy(session);
			ovr_Shutdown();
			SDL_GL_DeleteContext(glContext);
			SDL_DestroyWindow(window);
			SDL_Quit();
		}

		glBindFramebuffer(GL_READ_FRAMEBUFFER, mirrorFBOID);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		GLint w = winWidth;
		GLint h = winHeight;
		glBlitFramebuffer(0, h, w, 0,
			0, 0, w, h,
			GL_COLOR_BUFFER_BIT, GL_NEAREST);
		glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);
		SDL_GL_SwapWindow(window);




	}

	simxSetFloatSignal(clientID, "qdotECM_1", 0.0, simx_opmode_oneshot);
	simxSetFloatSignal(clientID, "qdotECM_2", 0.0, simx_opmode_oneshot);
	simxSetFloatSignal(clientID, "qdotECM_3", 0.0, simx_opmode_oneshot);
	simxSetFloatSignal(clientID, "qdotECM_4", 0.0, simx_opmode_oneshot);

	std::cout << "Ending Rift main loop\n";

	glDisableVertexAttribArray(Shader::ATTRIB_TEXTURE2D_POS);
	glDisableVertexAttribArray(Shader::ATTRIB_VERTICES_POS);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindTexture(GL_TEXTURE_2D, 0);
	glUseProgram(0);
	glBindVertexArray(0);
	glDeleteBuffers(3, rectVBO);
	ovr_DestroyTextureSwapChain(session, textureChain);
	ovr_DestroyMirrorTexture(session, mirrorTexture);
	ovr_Destroy(session);
	ovr_Shutdown();
	SDL_GL_DeleteContext(glContext);
	SDL_DestroyWindow(window);
	SDL_Quit();

	std::cout << "Ending Rift thread\n";
}

int main(int argc, char **argv) {

	std::cout << "Setting V-rep local host\n";

	clientID = simxStart("127.0.0.1", 19997, true, true, 5000, 5);

	std::cout << "Getting objects handles\n";

	while (simxGetObjectHandle(clientID, "LeftEye", &eyes[0], simx_opmode_blocking));
	while (simxGetObjectHandle(clientID, "RightEye", &eyes[1], simx_opmode_blocking));

	running = true;

	std::cout << "Starting V-rep simulation\n";

	simxStartSimulation(clientID, simx_opmode_blocking);
	simxGetFloatSignal(clientID, "Ts", &Ts, simx_opmode_blocking);

	std::cout << "Simulation time step: " << Ts << std::endl;

	// Init Logging
	initLog();

	// Init and launch threads
	boost::thread_group group;
	group.create_thread(boost::bind(escapeCallback));
	group.create_thread(boost::bind(hapticCHAICallback));
	group.create_thread(boost::bind(riftCallback));
	group.join_all();


	std::cout << "Ending V-rep simulation\n";

	std::cout << "Stopping simulation ... " << std::endl;
	simxStopSimulation(clientID, simx_opmode_blocking);
	simxFinish(clientID);

	// Save to file the logged data
	logData();


	// Close files for logging
	std::ofstream seqFileOut;
	seqFileOut.open("sequeceFile.txt", std::ios::out);
	seqFileOut << fileSeq;
	seqFileOut.close();

	return 0;
}


Eigen::Matrix3f eulAng2Rot(const float& alpha, const float& beta, const float& gamma) {

	Eigen::Matrix3f R, Rx, Ry, Rz;
	R.setIdentity();

	Rx << 1,          0,           0,
		  0, cos(alpha), -sin(alpha),
		  0, sin(alpha),  cos(alpha);
	
	Ry << cos(beta), 0, sin(beta),
				  0, 1,         0,
		 -sin(beta), 0, cos(beta);
	
	Rz << cos(gamma), -sin(gamma), 0,
		  sin(gamma),  cos(gamma), 0,
		           0,           0, 1;

	
	R = Rx * Ry * Rz;

	return R;
}



double getTime() {
	SYSTEMTIME tv;
	GetLocalTime(&tv);
	return tv2sec(tv);
}


void usleep(DWORD waitTime) {
	LARGE_INTEGER perfCnt, start, now;

	QueryPerformanceFrequency(&perfCnt);
	QueryPerformanceCounter(&start);

	do {
		QueryPerformanceCounter((LARGE_INTEGER*)&now);
	} while ((now.QuadPart - start.QuadPart) / float(perfCnt.QuadPart) * 1000 * 1000 < waitTime);
}


void initLog() {


	_getcwd(curDirAndFile, sizeof(curDirAndFile));
	path = std::string(curDirAndFile);
	path += "/plot";

	std::ifstream seqFileIn;
	seqFileIn.open("sequeceFile.txt", std::ios::in);

	if (seqFileIn.is_open())
	{
		seqFileIn >> fileSeq;
		fileSeq++;
	}
	else {
		fileSeq = 1;
	}

	path = path + std::to_string(fileSeq);

	path += "/";

	CreateDirectory(path.c_str(), NULL);

	// Oculus stringstream
	vtrSS.str("");				//!< 6D Velocity of the rift display expressed in the tracker reference frame
	vrrSS.str("");				//!< 6D Velocity of the rift display expressed in the rift reference frame
	Ttr.str("");					//!< Homogeneous transformation matrix expressing the pose of the Oculus rift wrt the tracker reference frame

											// ECM stringstream
	qdotECMSS.str("");			//!< 4x1 Vector of ECM joint velocities
	qECMSS.str("");				//!< 4x1 Vector of ECM joint position
	TbcSS.str("");				//!< Homogeneous transformation matrix expressing the pose of the left camera wrt the base frame of the ECM
	vcamLSS.str("");				//!< Velocity vector of the left camera of the ECM
	vcamRSS.str("");				//!< Velocity vector of the right camera of the ECM

											// Geomagic stringstream
	leftHipBaseVelSS.str("");		//!< 6D Velocity of the left hip with respect to the base frame of the left geomagic
	rightHipBaseVelSS.str("");	//!< 6D Velocity of the right hip with respect to the base frame of the right geomagic
	leftHipPSMVelSS.str("");		//!< 6D Velocity of the left hip with respect to the base frame of the Left PSM
	rightHipPSMVelSS.str("");	//!< 6D Velocity of the right hip with respect to the base frame of the Left PSM
	leftHipRotSS.str("");			//!< Rotation matrix expressing the orientation of the left hip wrt the base frame of the left geomagic
	rightHipRotSS.str("");		//!< Rotation matrix expressing the orientation of the left hip wrt the base frame of the right geomagic
	leftHipPosSS.str("");			//!< Position of the left hip wrt the base frame of the left geomagic
	rightHipPosSS.str("");		//!< Position of the right hip wrt the base frame of the right geomagic

	// PSM stringstream
	qdotPSM1SS.str("");			//!< 6x1 Vector of PSM1 joint velocity
	qdotPSM2SS.str("");			//!< 6x1 Vector of PSM1 joint position
	qPSM1SS.str("");				//!< 6x1 Vector of PSM2 joint velocity
	qPSM2SS.str("");				//!< 6x1 Vector of PSM2 joint position
	Tb1gSS.str("");				//!< Homogeneous transformation matrix expressing the pose of the left gripper (PSM1) wrt the base frame of the PSM1
	Tb2gSS.str("");				//!< Homogeneous transformation matrix expressing the pose of the right gripper (PSM2) wrt the base frame of the PSM2
	vb1gSS.str("");				//!< Velocity vector of the left gripper (PSM1) expressed wrt base frame of the PSM1
	vb2gSS.str("");				//!< Velocity vector of the right gripper (PSM2) expressed wrt base frame of the PSM2

	// Signals
	clutchButtonSS.str("");
	holdhButtonSS.str("");
	leftRaisingClutchEdgeSS.str("");
	rightRaisingClutchEdgeSS.str("");
	leftTrailingClutchEdgeSS.str("");
	rightTrailingClutchEdgeSS.str("");
	leftTriggerClutchButtonSS.str("");
	rightTriggerClutchButtonSS.str("");


}


void logData() {

	// Oculus stringstream
	std::string vtrFile = path + "vtr.txt";
	std::string vrrFile = path + "vrr.txt";
	std::string TtrFile = path + "Ttr.txt";
	
	// ECM stringstream
	std::string qdotECMFile = path + "qdotECM.txt";
	std::string qECMFile = path + "qECM.txt";
	std:string TbcFile = path + "Tbc.txt";
	std::string vcamLFile = path + "vcamL.txt";
	std::string vcamRSSFile = path + "vcamRSS.txt";
	
	// Geomagic stringstream
	std::string leftHipBaseVelFile = path + "leftHipBaseVel.txt";
	std::string rightHipBaseVelFile = path + "rightHipBaseVel.txt";
	std::string leftHipPSMVelFile = path + "leftHipPSMVel.txt";
	std::string rightHipPSMVelFile = path + "righttHipPSMVel.txt";
	std::string leftHipRotFile = path + "leftHipRot.txt";
	std::string rightHipRotFile = path + "rightHipRot.txt";
	std::string leftHipPosFile = path + "leftHipPos.txt";
	std::string rightHipPosFile = path + "rightHipPos.txt";
	
	// PSM stringstream
	std::string qdotPSM1File = path + "qdotPSM1.txt";
	std::string qdotPSM2File = path + "qdotPSM2.txt";
	std::string qPSM1File = path + "qPSM1.txt";
	std::string qPSM2File = path + "qPSM2.txt";
	std::string Tb1gFile = path + "Tb1g.txT";
	std::string Tb2gFile = path + "Tb2g.TXT";
	std::string vb1gFile = path + "vb1g.txt";
	std::string vb2gFile = path + "vb2g.txt";

	// Signals
	std::string clutchButtonFile = path + "clutchButton.txt";
	std::string holdhButtonFile = path + "holdhButton.txt";
	std::string leftRaisingClutchEdgeFile = path + "leftRaisingClutchEdge.txt";
	std::string rightRaisingClutchEdgeFile = path + "rightRaisingClutchEdge.txt";
	std::string leftTrailingClutchEdgeFile = path + "leftTrailingClutchEdge.txt";
	std::string rightTrailingClutchEdgeFile = path + "rightTrailingClutchEdge.txt";
	std::string leftTriggerClutchButtonFile = path + "leftTriggerClutchButton.txt";
	std::string rightTriggerClutchButtonFile = path + "rightTriggerClutchButton.txt";

	saveToFile(leftHipBaseVelSS, leftHipBaseVelFile);
	saveToFile(rightHipBaseVelSS, rightHipBaseVelFile);
	saveToFile(leftHipPSMVelSS, leftHipPSMVelFile);
	saveToFile(rightHipPSMVelSS, rightHipPSMVelFile);
	saveToFile(leftHipPosSS, leftHipPosFile);
	saveToFile(rightHipPosSS, rightHipPosFile);

	saveToFile(qPSM1SS, qPSM1File);
	saveToFile(qPSM2SS, qPSM2File);
	saveToFile(qdotPSM1SS, qdotPSM1File);
	saveToFile(qdotPSM2SS, qdotPSM2File);
	saveToFile(vb1gSS, vb1gFile);
	saveToFile(vb2gSS, vb2gFile);	

	saveToFile(qdotECMSS, qdotECMFile);
	saveToFile(qECMSS, qECMFile);
	saveToFile(vtrSS, vtrFile);
}


void saveToFile(std::stringstream& buff, std::string name)
{
	//---------------------------------------------
	//Input/Output State Bits
	//---------------------------------------------
	//| Goodbit | false/true | ERROS/NO ERRORS
	//| Eofbit  | 1          | End of File bit SET
	//| Failbit | 2          | Fail bit SET
	//| Badbit  | 4          | Bad bit SET
	//---------------------------------------------
	//| E+F     | 3
	//| E+B     | 5
	//| F+B     | 6
	//| E+F+B   | 7
	//---------------------------------------------
	std::ofstream data;
	data.open(name, std::ofstream::app);
	int i = 0;
	int j = 0;
	if (data.is_open())
	{
		std::cout << "Writing to " << name;
		do
		{
			if (i>3) break;
			else
			{
				i++;
				std::cout << ".";
				data.write(buff.str().c_str(), buff.str().length());
			}
		} while (!data.good());
		if (data.good())
		{
			std::cout << "OK";
			std::cout << "\nClosing file";
		}
		else
		{
			std::cout << "ERRORS ON STREAM(" << data.rdstate() << ")";
			std::cout << "\nClosing file";
			data.clear();
		}
		do
		{
			std::cout << ".";
			data.close();
		} while (data.fail());
		std::cout << "OK\n";
	}
	else
	{
		std::cout << "Error while opening file " << name << "!";
	}

}

